/// This interface is used to automatically override the hashCode and equals
/// functions.
abstract class Equality {
  /// The values which should be checked in hashCode and equals.
  get equalityValues;

  @override
  int get hashCode => EqualityComparator.hash(equalityValues);

  @override
  bool operator ==(other) => EqualityComparator.equals(this, other);
}

/// This class adds the ability to easily compare two objects.
class EqualityComparator {
  static const PRIME = 31;

  /// Generates the hash of the given object. If the given object is either a
  /// list or a map, a custom hash is generated because dart uses the memory
  /// pointer as hash code instead of looking at the content.
  static int hash(val) {
    if (val == null) {
      return 0;
    }

    if (val is List) {
      return _hashForList(val);
    } else if (val is Map) {
      return _hashForMap(val);
    } else if (val is Equality) {
      return hash(val.equalityValues);
    } else {
      return val.hashCode;
    }
  }

  /// Compares only the hash codes of the given objects.
  static bool quickEquals(a, b) => hash(a) == hash(b);

  /// Deeply checks whether object a equals object b. When lists or maps are
  /// given the content is checked.
  static bool equals(a, b) {
    int aHash = hash(a);
    int bHash = hash(b);

    if (aHash != bHash) {
      return false;
    }

    if (a == null) {
      return true;
    }

    if (a is List && b is List) {
      return _equalsForList(a, b);
    } else if (a is Map && b is Map) {
      return _equalsForMap(a, b);
    } else if (a is Equality && b is Equality) {
      return equals(a.equalityValues, b.equalityValues);
    } else {
      return a == b;
    }
  }

  /// Computes the hash code for all elements of a list
  static int _hashForList(List<dynamic> list) {
    int result = 0;

    for (dynamic val in list) {
      result = PRIME * result + hash(val);
    }

    return result;
  }

  /// Computes the hash code for all elements in a map
  static int _hashForMap(Map<dynamic, dynamic> map) {
    int result = 0;

    for (MapEntry<dynamic, dynamic> entry in map.entries) {
      result = PRIME * result + hash(entry.key);
      result = PRIME * result + hash(entry.value);
    }

    return result;
  }

  /// Checks whether the content of the given lists are equal
  static bool _equalsForList(Iterable<dynamic> a, Iterable<dynamic> b) {
    if (a.length != b.length) {
      return false;
    }

    for (int i = 0; i < a.length; i++) {
      if (!equals(a.elementAt(i), b.elementAt(i))) {
        return false;
      }
    }

    return true;
  }

  /// Checks whether the content of the given maps are equal
  static bool _equalsForMap(Map<dynamic, dynamic> a, Map<dynamic, dynamic> b) =>
      _equalsForList(a.keys, b.keys) && _equalsForList(a.values, b.values);
}
