/// A sorted list is a list which is getting sorted when inserting new elements.
/// It uses binary search to find the index where to insert the element.
class SortedList<T extends Comparable> extends Iterable<T> {
  List<T> _elements;

  SortedList() : _elements =  List();

  /// Performs a binary search against the elements and returns the  index where
  /// the item was found. When insert is set to true, the index where to insert
  /// the new element will be returned. Otherwise `null` is returned when no
  /// matching element was found.
  int binarySearch(var item, {Comparator comparator, bool insert = false}) {
    int left = 0;
    int right = _elements.length - 1;

    while (left <= right) {
      int middle = (left + right) ~/ 2;

      int comparison;
      if (comparator == null) {
        comparison = _elements[middle].compareTo(item);
      } else {
        comparison = comparator(_elements[middle], item);
      }

      if (comparison == -1) {
        left = middle + 1;
      } else if (comparison == 1) {
        right = middle - 1;
      } else {
        return middle;
      }
    }

    return (insert ? left : null);
  }

  /// Searches for a specific element in the list and returns the element if
  /// available.
  T find(item, {Comparator comparator}) {
    int index = binarySearch(item, comparator: comparator);
    return index == null ? null : _elements[index];
  }

  /// Returns the element stored at the specified index location
  T operator [](int index) => _elements[index];

  /// Returns the amount of elements in this list
  int get length => _elements.length;

  /// Inserts a new element into the list
  void add(T item) {
    int idx = binarySearch(item, insert: true);
    _elements.insert(idx, item);
  }

  /// Adds multiple items into the list
  void addAll(Iterable<T> iterable) {
    for (T item in iterable) {
      add(item);
    }
  }

  @override
  Iterator<T> get iterator => _elements.iterator;

  /// Returns the raw list of elements.
  ///
  /// **Note:** this should not be used to insert new elements since those will
  /// not be sorted automatically.
  List<T> get list => _elements;
}
